@@ -21,6 +21,7 @@
  *
  */
 
+#include "aconfig.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -33,6 +34,7 @@
 #include <syslog.h>
 #include <pthread.h>
 #include "alsaloop.h"
+#include "os_compat.h"
 
 #define XRUN_PROFILE_UNKNOWN (-10000000)
 
@@ -151,7 +153,7 @@ static int setparams_stream(struct loopback_handle *lhandle,
 #ifdef USE_SAMPLERATE
 	    !lhandle->loopback->src_enable &&
 #endif
-	    (int)rrate != lhandle->rate) {
+	    rrate != lhandle->rate) {
 		logit(LOG_CRIT, "Rate does not match (requested %uHz, got %uHz, resample %u)\n", lhandle->rate, rrate, lhandle->resample);
 		return -EINVAL;
 	}
@@ -623,7 +625,7 @@ static void buf_add_src(struct loopback *loop)
 	}
 }
 #else
-static void buf_add_src(struct loopback *loop)
+static void buf_add_src(struct loopback *loop ATTRIBUTE_UNUSED)
 {
 }
 #endif
@@ -645,13 +647,15 @@ static int xrun(struct loopback_handle *lhandle)
 	int err;
 
 	if (lhandle == lhandle->loopback->play) {
-		logit(LOG_DEBUG, "underrun for %s\n", lhandle->id);
+		if (verbose)
+			logit(LOG_DEBUG, "underrun for %s\n", lhandle->id);
 		xrun_stats(lhandle->loopback);
 		if ((err = snd_pcm_prepare(lhandle->handle)) < 0)
 			return err;
 		lhandle->xrun_pending = 1;
 	} else {
-		logit(LOG_DEBUG, "overrun for %s\n", lhandle->id);
+		if (verbose)
+			logit(LOG_DEBUG, "overrun for %s\n", lhandle->id);
 		xrun_stats(lhandle->loopback);
 		if ((err = snd_pcm_prepare(lhandle->handle)) < 0)
 			return err;
@@ -684,7 +688,7 @@ static int readit(struct loopback_handle *lhandle)
 		if ((err = suspend(lhandle)) < 0)
 			return err;
 	}
-	if (avail > buf_avail(lhandle)) {
+	if ((snd_pcm_uframes_t)avail > buf_avail(lhandle)) {
 		lhandle->buf_over += avail - buf_avail(lhandle);
 		avail = buf_avail(lhandle);
 	} else if (avail == 0) {
@@ -723,7 +727,7 @@ static int readit(struct loopback_handle *lhandle)
 			       r, lhandle->frame_size, lhandle->loopback->cfile);
 #endif
 		res += r;
-		if (lhandle->max < res)
+		if (lhandle->max < (snd_pcm_uframes_t)res)
 			lhandle->max = res;
 		lhandle->counter += r;
 		lhandle->buf_count += r;
@@ -802,9 +806,9 @@ static snd_pcm_sframes_t remove_samples(struct loopback *loop,
 	struct loopback_handle *capt = loop->capt;
 
 	if (loop->play->buf == loop->capt->buf) {
-		if (count > loop->play->buf_count)
+		if ((snd_pcm_uframes_t)count > loop->play->buf_count)
 			count = loop->play->buf_count;
-		if (count > loop->capt->buf_count)
+		if ((snd_pcm_uframes_t)count > loop->capt->buf_count)
 			count = loop->capt->buf_count;
 		capt->buf_count -= count;
 		play->buf_pos += count;
@@ -813,11 +817,11 @@ static snd_pcm_sframes_t remove_samples(struct loopback *loop,
 		return count;
 	}
 	if (capture_preferred) {
-		if (count > capt->buf_count)
+		if ((snd_pcm_uframes_t)count > capt->buf_count)
 			count = capt->buf_count;
 		capt->buf_count -= count;
 	} else {
-		if (count > play->buf_count)
+		if ((snd_pcm_uframes_t)count > play->buf_count)
 			count = play->buf_count;
 		play->buf_count -= count;
 	}
@@ -828,7 +832,7 @@ static int xrun_sync(struct loopback *loop)
 {
 	struct loopback_handle *play = loop->play;
 	struct loopback_handle *capt = loop->capt;
-	snd_pcm_uframes_t fill = get_whole_latency(loop);
+	snd_pcm_sframes_t fill = get_whole_latency(loop);
 	snd_pcm_sframes_t pdelay, cdelay, delay1, pdelay1, cdelay1, diff;
 	int err;
 
@@ -928,7 +932,7 @@ static int xrun_sync(struct loopback *loop)
 	}
 	if (delay1 > fill) {
 		diff = (delay1 - fill) / play->pitch;
-		if (diff > play->buf_count)
+		if ((snd_pcm_uframes_t)diff > play->buf_count)
 			diff = play->buf_count;
 		if (verbose > 6)
 			snd_output_printf(loop->output,
@@ -943,7 +947,7 @@ static int xrun_sync(struct loopback *loop)
 	}
 	if (delay1 > fill) {
 		diff = (delay1 - fill) / capt->pitch;
-		if (diff > capt->buf_count)
+		if ((snd_pcm_uframes_t)diff > capt->buf_count)
 			diff = capt->buf_count;
 		if (verbose > 6)
 			snd_output_printf(loop->output,
@@ -962,7 +966,7 @@ static int xrun_sync(struct loopback *loop)
 		if (verbose > 6)
 			snd_output_printf(loop->output,
 				"sync: xrun_pending, silence filling %li / buf_count=%li\n", (long)diff, play->buf_count);
-		if (fill > delay1 && play->buf_count < diff) {
+		if (fill > delay1 && play->buf_count < (snd_pcm_uframes_t)diff) {
 			diff = diff - play->buf_count;
 			if (verbose > 6)
 				snd_output_printf(loop->output,
@@ -993,7 +997,7 @@ static int xrun_sync(struct loopback *loop)
 			logit(LOG_CRIT, "%s start failed: %s\n", play->id, snd_strerror(err));
 			return err;
 		}
-	} else if (delay1 < fill) {
+	} else if (delay1 < (snd_pcm_sframes_t)fill) {
 		diff = (fill - delay1) / play->pitch;
 		while (diff > 0) {
 			delay1 = play->buf_size - play->buf_pos;
@@ -1060,10 +1064,16 @@ static int set_rate_shift(struct loopback_handle *lhandle, double pitch)
 {
 	int err;
 
-	if (lhandle->ctl_rate_shift == NULL)
+	if (lhandle->ctl_rate_shift) {
+		snd_ctl_elem_value_set_integer(lhandle->ctl_rate_shift, 0, pitch * 100000);
+		err = snd_ctl_elem_write(lhandle->ctl, lhandle->ctl_rate_shift);
+	} else if (lhandle->ctl_pitch) {
+		// 'Playback/Capture Pitch 1000000' requires reciprocal to pitch
+		snd_ctl_elem_value_set_integer(lhandle->ctl_pitch, 0, (1 / pitch) * 1000000);
+		err = snd_ctl_elem_write(lhandle->ctl, lhandle->ctl_pitch);
+	} else {
 		return 0;
-	snd_ctl_elem_value_set_integer(lhandle->ctl_rate_shift, 0, pitch * 100000);
-	err = snd_ctl_elem_write(lhandle->ctl, lhandle->ctl_rate_shift);
+	}
 	if (err < 0) {
 		logit(LOG_CRIT, "Cannot set PCM Rate Shift element for %s: %s\n", lhandle->id, snd_strerror(err));
 		return err;
@@ -1096,7 +1106,8 @@ void update_pitch(struct loopback *loop)
 #endif
 	}
 	else if (loop->sync == SYNC_TYPE_PLAYRATESHIFT) {
-		set_rate_shift(loop->play, pitch);
+		// pitch is capture-based, playback side requires reciprocal
+		set_rate_shift(loop->play, 1 / pitch);
 #ifdef USE_SAMPLERATE
 		if (loop->use_samplerate) {
 			loop->src_data.src_ratio = 
@@ -1167,27 +1178,57 @@ static int get_channels(struct loopback_handle *lhandle)
 	return snd_ctl_elem_value_get_integer(lhandle->ctl_channels, 0);
 }
 
-static void openctl_elem(struct loopback_handle *lhandle,
-			 int device, int subdevice,
-			 const char *name,
-			 snd_ctl_elem_value_t **elem)
+static int openctl_elem_id(struct loopback_handle *lhandle, snd_ctl_elem_id_t *id,
+		snd_ctl_elem_value_t **elem)
 {
 	int err;
 
 	if (snd_ctl_elem_value_malloc(elem) < 0) {
 		*elem = NULL;
-	} else {
-		snd_ctl_elem_value_set_interface(*elem,
-						 SND_CTL_ELEM_IFACE_PCM);
-		snd_ctl_elem_value_set_device(*elem, device);
-		snd_ctl_elem_value_set_subdevice(*elem, subdevice);
-		snd_ctl_elem_value_set_name(*elem, name);
-		err = snd_ctl_elem_read(lhandle->ctl, *elem);
-		if (err < 0) {
-			snd_ctl_elem_value_free(*elem);
-			*elem = NULL;
-		}
+		return -ENOMEM;
 	}
+	snd_ctl_elem_value_set_id(*elem, id);
+	snd_ctl_elem_value_set_interface(*elem, SND_CTL_ELEM_IFACE_PCM);
+	err = snd_ctl_elem_read(lhandle->ctl, *elem);
+	if (err < 0) {
+		snd_ctl_elem_value_free(*elem);
+		*elem = NULL;
+		return err;
+	} else {
+		snd_output_printf(lhandle->loopback->output,
+				"Opened PCM element %s of %s, device %d, subdevice %d\n",
+				snd_ctl_elem_id_get_name(id), snd_ctl_name(lhandle->ctl),
+				snd_ctl_elem_id_get_device(id),
+				snd_ctl_elem_id_get_subdevice(id));
+		return 0;
+	}
+}
+
+static int openctl_elem(struct loopback_handle *lhandle,
+			 int device, int subdevice,
+			 const char *name,
+			 snd_ctl_elem_value_t **elem)
+{
+	snd_ctl_elem_id_t *id;
+
+	snd_ctl_elem_id_alloca(&id);
+	snd_ctl_elem_id_set_device(id, device);
+	snd_ctl_elem_id_set_subdevice(id, subdevice);
+	snd_ctl_elem_id_set_name(id, name);
+	return openctl_elem_id(lhandle, id, elem);
+}
+
+static int openctl_elem_ascii(struct loopback_handle *lhandle, char *ascii_name,
+		snd_ctl_elem_value_t **elem)
+{
+	snd_ctl_elem_id_t *id;
+
+	snd_ctl_elem_id_alloca(&id);
+	if (snd_ctl_ascii_elem_id_parse(id, ascii_name)) {
+		fprintf(stderr, "Wrong control identifier: %s\n", ascii_name);
+		return -EINVAL;
+	}
+	return openctl_elem_id(lhandle, id, elem);
 }
 
 static int openctl(struct loopback_handle *lhandle, int device, int subdevice)
@@ -1196,14 +1237,30 @@ static int openctl(struct loopback_handle *lhandle, int device, int subdevice)
 
 	lhandle->ctl_rate_shift = NULL;
 	if (lhandle->loopback->play == lhandle) {
+		// play only
+		if (lhandle->prateshift_name) {
+			err = openctl_elem_ascii(lhandle, lhandle->prateshift_name,
+					&lhandle->ctl_rate_shift);
+			if (err < 0) {
+				logit(LOG_CRIT, "Unable to open playback PCM Rate Shift elem '%s'.\n",
+						lhandle->prateshift_name);
+				exit(EXIT_FAILURE);
+			}
+		} else
+			openctl_elem(lhandle, device, subdevice, "Playback Pitch 1000000",
+					&lhandle->ctl_pitch);
+		set_rate_shift(lhandle, 1);
 		if (lhandle->loopback->controls)
 			goto __events;
 		return 0;
 	}
+	// capture only
 	openctl_elem(lhandle, device, subdevice, "PCM Notify",
 			&lhandle->ctl_notify);
 	openctl_elem(lhandle, device, subdevice, "PCM Rate Shift 100000",
 			&lhandle->ctl_rate_shift);
+	openctl_elem(lhandle, device, subdevice, "Capture Pitch 1000000",
+			&lhandle->ctl_pitch);
 	set_rate_shift(lhandle, 1);
 	openctl_elem(lhandle, device, subdevice, "PCM Slave Active",
 			&lhandle->ctl_active);
@@ -1289,6 +1346,9 @@ static int closeit(struct loopback_handle *lhandle)
 	if (lhandle->ctl_rate_shift)
 		snd_ctl_elem_value_free(lhandle->ctl_rate_shift);
 	lhandle->ctl_rate_shift = NULL;
+	if (lhandle->ctl_pitch)
+		snd_ctl_elem_value_free(lhandle->ctl_pitch);
+	lhandle->ctl_pitch = NULL;
 	if (lhandle->ctl)
 		err = snd_ctl_close(lhandle->ctl);
 	lhandle->ctl = NULL;
@@ -1334,9 +1394,9 @@ int pcmjob_init(struct loopback *loop)
 	snprintf(id, sizeof(id), "%s/%s", loop->play->id, loop->capt->id);
 	id[sizeof(id)-1] = '\0';
 	loop->id = strdup(id);
-	if (loop->sync == SYNC_TYPE_AUTO && loop->capt->ctl_rate_shift)
+	if (loop->sync == SYNC_TYPE_AUTO && (loop->capt->ctl_rate_shift || loop->capt->ctl_pitch))
 		loop->sync = SYNC_TYPE_CAPTRATESHIFT;
-	if (loop->sync == SYNC_TYPE_AUTO && loop->play->ctl_rate_shift)
+	if (loop->sync == SYNC_TYPE_AUTO && (loop->play->ctl_rate_shift || loop->play->ctl_pitch))
 		loop->sync = SYNC_TYPE_PLAYRATESHIFT;
 #ifdef USE_SAMPLERATE
 	if (loop->sync == SYNC_TYPE_AUTO && loop->src_enable)
@@ -1615,7 +1675,7 @@ __again:
 		snd_output_printf(loop->output, "%s: silence queued %i samples\n", loop->id, err);
 	if (count > loop->play->buffer_size)
 		count = loop->play->buffer_size;
-	if (err != count) {
+	if (err != (int)count) {
 		logit(LOG_CRIT, "%s: initial playback fill error (%i/%i/%u)\n", loop->id, err, (int)count, loop->play->buffer_size);
 		err = -EIO;
 		goto __error;
@@ -1736,7 +1796,7 @@ static int ctl_event_check(snd_ctl_elem_value_t *val, snd_ctl_event_t *ev)
 }
 
 static int handle_ctl_events(struct loopback_handle *lhandle,
-			     unsigned short events)
+			     unsigned short events ATTRIBUTE_UNUSED)
 {
 	struct loopback *loop = lhandle->loopback;
 	snd_ctl_event_t *ev;
@@ -1761,12 +1821,12 @@ static int handle_ctl_events(struct loopback_handle *lhandle,
 			continue;
 		} else if (ctl_event_check(lhandle->ctl_rate, ev)) {
 			err = get_rate(lhandle);
-			if (lhandle->rate != err)
+			if ((int)lhandle->rate != err)
 				restart = 1;
 			continue;
 		} else if (ctl_event_check(lhandle->ctl_channels, ev)) {
 			err = get_channels(lhandle);
-			if (lhandle->channels != err)
+			if ((int)lhandle->channels != err)
 				restart = 1;
 			continue;
 		}
@@ -1905,7 +1965,7 @@ int pcmjob_pollfds_handle(struct loopback *loop, struct pollfd *fds)
 		if (play->xrun_pending || loop->reinit)
 			break;
 		loopcount++;
-	} while ((ccount > 0 || pcount > 0) && loopcount > 10);
+	} while ((ccount > 0 || pcount > 0) && loopcount < 10);
 	if (play->xrun_pending || capt->xrun_pending) {
 		if ((err = xrun_sync(loop)) < 0)
 			return err;
